@@ -7,7 +7,7 @@
 
 # %%
 import elli
-import elli.plot.structure as elliplot
+import elli.plot as elliplot
 import matplotlib.pyplot as plt
 import numpy as np
 from scipy.constants import c, pi
