@@ -12,10 +12,6 @@ from .structure import *
 from .utils import *
 from .spectraray import *
 from .materials_db import *
-from .fitting.params_hist import ParamsHist
-from .fitting.decorator_psi_delta import fit
-from .fitting.decorator_mmatrix import fit_mueller_matrix
-from .plot.mueller_matrix import plot_mmatrix
 
 if sys.version_info.major >= 3 and sys.version_info.minor > 7:
     from importlib.metadata import PackageNotFoundError, version
